﻿' See "Firefly Algorithms for Multimodal Optimization",
' X. Yang, 5th Intenational Symposium on Stochastic Algorithms, 2009.

Namespace FireflyAlgorithm
	Friend Class FireflyProgram
		Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin firefly algorithm optimization demo" & vbCrLf)
            Console.WriteLine("Goal is to solve the Michalewicz benchmark function")
			Console.WriteLine("The function has a known minimum value of -4.687658")
			Console.WriteLine("x = 2.2029 1.5707 1.2850 1.9231 1.7205")

			Dim numFireflies As Integer = 40 ' typically 15-40
			Dim [dim] As Integer = 5
			Dim maxEpochs As Integer = 1000
			Dim seed As Integer = 0
            'double errThresh = 1.0E-5 ' early exit

            Console.WriteLine(vbCrLf & "Setting numFireflies        = " & numFireflies)
            Console.WriteLine("Setting problem dim         = " & [dim])
			Console.WriteLine("Setting maxEpochs           = " & maxEpochs)
			Console.WriteLine("Setting initialization seed = " & seed)
            'Console.WriteLine("Setting early-exit error    = " + errThresh.ToString("F5"))

            Console.WriteLine(vbCrLf & "Starting firefly algorithm" & vbCrLf)
            Dim bestPosition() As Double = Solve(numFireflies, [dim], seed, maxEpochs)
            Console.WriteLine(vbCrLf & "Finished" & vbCrLf)

            Console.WriteLine("Best solution found: ")
			Console.Write("x = ")
			ShowVector(bestPosition, 4, True)
			Dim z As Double = Michalewicz(bestPosition)
			Console.Write("Value of function at best position = ")
			Console.WriteLine(z.ToString("F6"))
			Dim [error] As Double = FireflyProgram.Error(bestPosition)
			Console.Write("Error at best position = ")
			Console.WriteLine([error].ToString("F4"))

            Console.WriteLine(vbCrLf & "End firefly demo" & vbCrLf)
            Console.ReadLine()
		End Sub ' Main

		Private Shared Sub ShowVector(ByVal v() As Double, ByVal dec As Integer, ByVal nl As Boolean)
			For i As Integer = 0 To v.Length - 1
				Console.Write(v(i).ToString("F" & dec) & " ")
			Next i
			If nl = True Then
				Console.WriteLine("")
			End If
		End Sub

		Private Shared Function Solve(ByVal numFireflies As Integer, ByVal [dim] As Integer, ByVal seed As Integer, ByVal maxEpochs As Integer) As Double()
			Dim rnd As New Random(seed)
			Dim minX As Double = 0.0 ' specific to Michalewicz function
			Dim maxX As Double = 3.2

			Dim B0 As Double = 1.0 ' beta (attractiveness base)
            'double betaMin = 0.20
            Dim g As Double = 1.0 ' gamma (absorption for attraction)
			Dim a As Double = 0.20 ' alpha
			'double a0 = 1.0;    // base alpha for decay
			Dim displayInterval As Integer = maxEpochs \ 10

			Dim bestError As Double = Double.MaxValue
			Dim bestPosition([dim] - 1) As Double ' best ever

			Dim swarm(numFireflies - 1) As Firefly ' all null

			' initialize swarm at random positions
			For i As Integer = 0 To numFireflies - 1
				swarm(i) = New Firefly([dim]) ' position 0, error and intensity 0.0
				For k As Integer = 0 To [dim] - 1 ' random position
					swarm(i).position(k) = (maxX - minX) * rnd.NextDouble() + minX
				Next k
				swarm(i).error = [Error](swarm(i).position) ' associated error
				swarm(i).intensity = 1 / (swarm(i).error + 1) ' +1 prevent div by 0
				If swarm(i).error < bestError Then
					bestError = swarm(i).error
					For k As Integer = 0 To [dim] - 1
						bestPosition(k) = swarm(i).position(k)
					Next k
				End If
			Next i

			Dim epoch As Integer = 0
			Do While epoch < maxEpochs ' main processing
                'if (bestError < errThresh) break 'are we good?
                If epoch Mod displayInterval = 0 AndAlso epoch < maxEpochs Then ' show progress?
					Dim sEpoch As String = epoch.ToString().PadLeft(6)
					Console.Write("epoch = " & sEpoch)
					Console.WriteLine("   error = " & bestError.ToString("F14"))
				End If

				For i As Integer = 0 To numFireflies - 1 ' each firefly
					For j As Integer = 0 To numFireflies - 1 ' each other firefly. weird!
						If swarm(i).intensity < swarm(j).intensity Then
							' curr firefly i is less intense (i is worse) so move i toward j
							Dim r As Double = Distance(swarm(i).position, swarm(j).position)
							Dim beta As Double = B0 * Math.Exp(-g * r * r) ' original
                            'double beta = (B0 - betaMin) * Math.Exp(-g * r * r) + betaMin ' better
                            'double a = a0 * Math.Pow(0.98, epoch) ' better
                            For k As Integer = 0 To [dim] - 1
								swarm(i).position(k) += beta * (swarm(j).position(k) - swarm(i).position(k))
								swarm(i).position(k) += a * (rnd.NextDouble() - 0.5)
								If swarm(i).position(k) < minX Then
									swarm(i).position(k) = (maxX - minX) * rnd.NextDouble() + minX
								End If
								If swarm(i).position(k) > maxX Then
									swarm(i).position(k) = (maxX - minX) * rnd.NextDouble() + minX
								End If
							Next k
							swarm(i).error = [Error](swarm(i).position)
							swarm(i).intensity = 1 / (swarm(i).error + 1)
						End If
					Next j ' j
				Next i ' i each firefly

				Array.Sort(swarm) ' low error to high
				If swarm(0).error < bestError Then ' new best?
					bestError = swarm(0).error
					For k As Integer = 0 To [dim] - 1
						bestPosition(k) = swarm(0).position(k)
					Next k
				End If
				epoch += 1
			Loop ' while
			Return bestPosition
		End Function ' Solve

		Private Shared Function Distance(ByVal posA() As Double, ByVal posB() As Double) As Double
			Dim ssd As Double = 0.0 ' sum squared diffrences (Euclidean)
			For i As Integer = 0 To posA.Length - 1
				ssd += (posA(i) - posB(i)) * (posA(i) - posB(i))
			Next i
			Return Math.Sqrt(ssd)
		End Function

		Private Shared Function Michalewicz(ByVal xValues() As Double) As Double
			Dim result As Double = 0.0
			For i As Integer = 0 To xValues.Length - 1
				Dim a As Double = Math.Sin(xValues(i))
				Dim b As Double = Math.Sin(((i + 1) * xValues(i) * xValues(i)) / Math.PI)
				Dim c As Double = Math.Pow(b, 20)
				result += a * c
			Next i
			Return -1.0 * result
		End Function ' Michalewicz

		Private Shared Function [Error](ByVal xValues() As Double) As Double
			Dim [dim] As Integer = xValues.Length
			Dim trueMin As Double = 0.0
			If [dim] = 2 Then
				trueMin = -1.8013 ' approx.
			ElseIf [dim] = 5 Then
				trueMin = -4.687658 ' approx.
			ElseIf [dim] = 10 Then
				trueMin = -9.66015 ' approx.
			End If
			Dim calculated As Double = Michalewicz(xValues)
			Return (trueMin - calculated) * (trueMin - calculated)
		End Function
	End Class ' Program

	Public Class Firefly
		Implements IComparable(Of Firefly)
		Public position() As Double
		Public [error] As Double
		Public intensity As Double

		Public Sub New(ByVal [dim] As Integer)
			Me.position = New Double([dim] - 1){}
			Me.error = 0.0
			Me.intensity = 0.0
		End Sub

		Public Function CompareTo(ByVal other As Firefly) As Integer Implements IComparable(Of Firefly).CompareTo
			' allow auto sort low error to high
			If Me.error < other.error Then
				Return -1
			ElseIf Me.error > other.error Then
                Return 1
            Else
				Return 0
			End If
		End Function
	End Class ' class Firefly
End Namespace ' ns
